@@ -8,6 +8,7 @@ import os
 import time
 import errno
 import socket
+import abc
 
 import ctypes
 import ctypes.util
@@ -34,10 +35,12 @@ from tarantool.request import (
     RequestSubscribe,
     RequestUpdate,
     RequestUpsert,
-    RequestAuthenticate
+    RequestAuthenticate,
+    RequestExecute
 )
 from tarantool.space import Space
 from tarantool.const import (
+    CONNECTION_TIMEOUT,
     SOCKET_TIMEOUT,
     RECONNECT_MAX_ATTEMPTS,
     RECONNECT_DELAY,
@@ -48,12 +51,21 @@ from tarantool.const import (
     ITERATOR_ALL
 )
 from tarantool.error import (
+    Error,
     NetworkError,
     DatabaseError,
     InterfaceError,
+    ConfigurationError,
     SchemaError,
     NetworkWarning,
+    OperationalError,
+    DataError,
+    IntegrityError,
+    InternalError,
+    ProgrammingError,
+    NotSupportedError,
     SchemaReloadException,
+    Warning,
     warn
 )
 from tarantool.schema import Schema
@@ -65,8 +77,92 @@ from tarantool.utils import (
     ENCODING_DEFAULT,
 )
 
+# Based on https://realpython.com/python-interface/
+class ConnectionInterface(metaclass=abc.ABCMeta):
+    @classmethod
+    def __subclasshook__(cls, subclass):
+        return (hasattr(subclass, 'close') and
+                callable(subclass.close) and
+                hasattr(subclass, 'is_closed') and
+                callable(subclass.is_closed) and
+                hasattr(subclass, 'connect') and
+                callable(subclass.connect) and
+                hasattr(subclass, 'call') and
+                callable(subclass.call) and
+                hasattr(subclass, 'eval') and
+                callable(subclass.eval) and
+                hasattr(subclass, 'replace') and
+                callable(subclass.replace) and
+                hasattr(subclass, 'insert') and
+                callable(subclass.insert) and
+                hasattr(subclass, 'delete') and
+                callable(subclass.delete) and
+                hasattr(subclass, 'upsert') and
+                callable(subclass.upsert) and
+                hasattr(subclass, 'update') and
+                callable(subclass.update) and
+                hasattr(subclass, 'ping') and
+                callable(subclass.ping) and
+                hasattr(subclass, 'select') and
+                callable(subclass.select) and
+                hasattr(subclass, 'execute') and
+                callable(subclass.execute) or
+                NotImplemented)
 
-class Connection(object):
+    @abc.abstractmethod
+    def close(self):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def is_closed(self):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def connect(self):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def call(self, func_name, *args, **kwargs):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def eval(self, expr, *args, **kwargs):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def replace(self, space_name, values):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def insert(self, space_name, values):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def delete(self, space_name, key, **kwargs):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def upsert(self, space_name, tuple_value, op_list, **kwargs):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def update(self, space_name, key, op_list, **kwargs):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def ping(self, notime):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def select(self, space_name, key, **kwargs):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def execute(self, query, params, **kwargs):
+        raise NotImplementedError
+
+
+class Connection(ConnectionInterface):
     '''
     Represents connection to the Tarantool server.
 
@@ -75,11 +171,20 @@ class Connection(object):
     Also this class provides low-level interface to data manipulation
     (insert/delete/update/select).
     '''
-    Error = tarantool.error
+    # DBAPI Extension: supply exceptions as attributes on the connection
+    Error = Error
     DatabaseError = DatabaseError
     InterfaceError = InterfaceError
+    ConfigurationError = ConfigurationError
     SchemaError = SchemaError
     NetworkError = NetworkError
+    Warning = Warning
+    DataError = DataError
+    OperationalError = OperationalError
+    IntegrityError = IntegrityError
+    InternalError = InternalError
+    ProgrammingError = ProgrammingError
+    NotSupportedError = NotSupportedError
 
     def __init__(self, host, port,
                  user=None,
@@ -89,7 +194,9 @@ class Connection(object):
                  reconnect_delay=RECONNECT_DELAY,
                  connect_now=True,
                  encoding=ENCODING_DEFAULT,
-                 call_16=False):
+                 use_list=True,
+                 call_16=False,
+                 connection_timeout=CONNECTION_TIMEOUT):
         '''
         Initialize a connection to the server.
 
@@ -99,6 +206,11 @@ class Connection(object):
         creates network connection.
                              if False than you have to call connect() manualy.
         '''
+
+        if msgpack.version >= (1, 0, 0) and encoding not in (None, 'utf-8'):
+            raise ConfigurationError("Only None and 'utf-8' encoding option " +
+                                     "values are supported with msgpack>=1.0.0")
+
         if os.name == 'nt':
             libc = ctypes.WinDLL(
                 ctypes.util.find_library('Ws2_32'), use_last_error=True
@@ -120,10 +232,11 @@ class Connection(object):
         self.schema_version = 1
         self._socket = None
         self.connected = False
-        self.inconnect = False
         self.error = True
         self.encoding = encoding
+        self.use_list = use_list
         self.call_16 = call_16
+        self.connection_timeout = connection_timeout
         if connect_now:
             self.connect()
 
@@ -134,6 +247,13 @@ class Connection(object):
         self._socket.close()
         self._socket = None
 
+    def is_closed(self):
+        '''
+        Returns the state of the Connection instance
+        :rtype: Boolean
+        '''
+        return self._socket is None
+
     def connect_basic(self):
         if self.host == None:
             self.connect_unix()
@@ -151,7 +271,10 @@ class Connection(object):
             self.connected = True
             if self._socket:
                 self._socket.close()
-            self._socket = socket.create_connection((self.host, self.port))
+            self._socket = socket.create_connection(
+                (self.host, self.port), timeout=self.connection_timeout)
+            self._socket.settimeout(self.socket_timeout)
+            self._socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
         except socket.error as e:
             self.connected = False
             raise NetworkError(e)
@@ -168,7 +291,9 @@ class Connection(object):
             if self._socket:
                 self._socket.close()
             self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+            self._socket.settimeout(self.connection_timeout)
             self._socket.connect(self.port)
+            self._socket.settimeout(self.socket_timeout)
         except socket.error as e:
             self.connected = False
             raise NetworkError(e)
@@ -193,19 +318,11 @@ class Connection(object):
         :raise: `NetworkError`
         '''
         try:
-            self.inconnect = True
             self.connect_basic()
             self.handshake()
-            # It is important to set socket timeout *after* connection.
-            # Otherwise the timeout exception will be raised, even when
-            # the connection fails because the server is simply
-            # not bound to port
-            self._socket.settimeout(self.socket_timeout)
             self.load_schema()
-            self.inconnect = False
-        except socket.error as e:
+        except Exception as e:
             self.connected = False
-            self.inconnect = False
             raise NetworkError(e)
 
     def _recv(self, to_read):
@@ -251,7 +368,7 @@ class Connection(object):
 
     def _send_request_wo_reconnect(self, request):
         '''
-        :rtype: `Response` instance
+        :rtype: `Response` instance or subclass
 
         :raise: NetworkError
         '''
@@ -261,7 +378,7 @@ class Connection(object):
         while True:
             try:
                 self._socket.sendall(bytes(request))
-                response = Response(self, self._read_response())
+                response = request.response_class(self, self._read_response())
                 break
             except SchemaReloadException as e:
                 self.update_schema(e.schema_version)
@@ -274,8 +391,6 @@ class Connection(object):
         Check that connection is alive using low-level recv from libc(ctypes)
         **Due to bug in python - timeout is internal python construction.
         '''
-        if self.inconnect:
-            return
         if not self._socket:
             return self.connect()
 
@@ -332,13 +447,7 @@ class Connection(object):
                 raise NetworkError(
                     socket.error(last_errno, errno.errorcode[last_errno]))
             attempt += 1
-
         self.handshake()
-        # It is important to set socket timeout *after* connection.
-        # Otherwise the timeout exception will be raised, even when
-        # the connection fails because the server is simply
-        # not bound to port
-        self._socket.settimeout(self.socket_timeout)
 
     def _send_request(self, request):
         '''
@@ -794,3 +903,36 @@ class Connection(object):
         Need override for async io connection
         '''
         return 0
+
+    def execute(self, query, params=None):
+        '''
+        Execute SQL request.
+
+        Tarantool binary protocol for SQL requests
+        supports "qmark" and "named" param styles.
+        Sequence of values can be used for "qmark" style.
+        A mapping is used for "named" param style
+        without leading colon in the keys.
+
+        Example for "qmark" arguments:
+        >>> args = ['email@example.com']
+        >>> c.execute('select * from "users" where "email"=?', args)
+
+        Example for "named" arguments:
+        >>> args = {'email': 'email@example.com'}
+        >>> c.execute('select * from "users" where "email"=:email', args)
+
+        :param query: SQL syntax query
+        :type query: str
+
+        :param params: Bind values to use in the query.
+        :type params: list, dict
+
+        :return: query result data
+        :rtype: `Response` instance
+        '''
+        if not params:
+            params = []
+        request = RequestExecute(self, query, params)
+        response = self._send_request(request)
+        return response
